-- =============================================================================
-- Copyright © 2017-2018 by Cryptographic Engineering Research Group (CERG),
-- ECE Department, George Mason University
-- Fairfax, VA, U.S.A.
-- Author: Farnoud Farahmand
-- =============================================================================
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
library work;
use work.ntru_pkg.all;

entity bwc2 is
    generic(
        W    : integer := 32
    );
    port(
        clk        : in std_logic;
        rst        : in std_logic;
        done       : out std_logic;
        chk_failed : out std_logic;
        --! Input FIFO ports
        infifo_dout  : in std_logic_vector(W-1 downto 0);
        infifo_read  : out std_logic;
        infifo_empty : in std_logic;
        --! Output handshaking ports
        output_data  : out std_logic_vector(2*2*W -1 downto 0);
        output_valid : out std_logic;
        output_ready : in std_logic
    );
end bwc2;

architecture arch of bwc2 is
    --! constants
    constant piso_width : integer := 8;

    --! Datapath signals
    signal piso_out     : std_logic_vector(piso_width-1 downto 0);
    signal o2t_out      : std_logic_vector(9 downto 0);
    signal r640 	    : std_logic_vector(639 downto 0);

    --! Controler signals
    type t_state is (S_RESET, S_FIRST_LOAD, S_SHIFT_LOAD, S_WAIT);
    signal state, state_next       : t_state;
    signal checkpoint, checkpoint_next : std_logic_vector(2 downto 0);
    signal c_piso_next, c_piso_r       : std_logic_vector(1 downto 0);
    signal c_sipo_next, c_sipo_r       : std_logic_vector(5 downto 0);
    signal c_Wcoef_next, c_Wcoef_r     : std_logic_vector(8 downto 0);
    signal c_Rcoef_next, c_Rcoef_r     : std_logic_vector(8 downto 0);
    signal c_hashw8_next, c_hashw8_r   : std_logic_vector(8 downto 0);
    signal c_hashw32_next, c_hashw32_r : std_logic_vector(6 downto 0);
    signal sipo_full_next, sipo_full_r : std_logic;
    signal wait_next, wait_r           : std_logic;
    signal done_next, done_r           : std_logic;
    signal bypass_empty_next, bypass_empty_r : std_logic; --! should be changed
    signal chk_failed_next, chk_failed_r     : std_logic;
    signal hash_done_next, hash_done_r : std_logic;
    signal en_piso       : std_logic;
    signal ld_piso       : std_logic;
    signal en_sipo       : std_logic;
    signal sel_o         : std_logic_vector(2 downto 0);
    signal lseq_243      : std_logic;
    signal bypass_empty  : std_logic;

    --! FIFO signals
    signal fifo_full     : std_logic;
    signal fifo_write    : std_logic;
    signal fifo_empty    : std_logic;
    signal fifo_empty_orig : std_logic;
    signal fifo_read     : std_logic;
    signal fifo_dout     : std_logic_vector(9 downto 0);

    signal NUM_HASH32_REQ_MGF   : integer;

begin

--! CONTROLLER
p_fsm: process(clk)
begin
    if rising_edge(clk) then
        if (rst = '1') then
            state      <= S_RESET;
            checkpoint <= "000";
        else
            state      <= state_next;
            checkpoint <= checkpoint_next;
        end if;
        c_piso_r    <= c_piso_next;
        c_sipo_r    <= c_sipo_next;
        c_Wcoef_r   <= c_Wcoef_next;
        c_Rcoef_r   <= c_Rcoef_next;
        c_hashw8_r  <= c_hashw8_next;
        c_hashw32_r <= c_hashw32_next;
        wait_r      <= wait_next;
        sipo_full_r <= sipo_full_next;
        bypass_empty_r <= bypass_empty_next;
        done_r         <= done_next;
        chk_failed_r   <= chk_failed_next;
        hash_done_r    <= hash_done_next;
    end if;
end process;

p_comb: process(state, checkpoint, infifo_empty, c_piso_r,
    c_sipo_r, lseq_243, output_ready, fifo_full, wait_r, sipo_full_r, fifo_empty,
    c_Wcoef_r, c_Rcoef_r, c_hashw8_r, done_r, chk_failed_r, fifo_read,
    c_hashw32_r, hash_done_r, NUM_HASH32_REQ_MGF)

begin
    --! Default values
    state_next      <= state;
    checkpoint_next <= checkpoint;
    c_piso_next     <= c_piso_r;
    c_sipo_next     <= c_sipo_r;
    c_Wcoef_next    <= c_Wcoef_r;
    c_Rcoef_next    <= c_Rcoef_r;
    c_hashw8_next   <= c_hashw8_r;
    c_hashw32_next   <= c_hashw32_r;
    wait_next       <= wait_r;
    sipo_full_next  <= sipo_full_r;
    bypass_empty_next <= bypass_empty_r;
    done_next       <= done_r;
    chk_failed_next <= chk_failed_r;
    hash_done_next  <= hash_done_r;
    sel_o           <= "101";
    en_piso         <= '0';
    ld_piso         <= '0';
    en_sipo         <= '0';
    infifo_read     <= '0';
    output_valid    <= '0';
    --! FIFO
    fifo_write      <= '0';
    fifo_read       <= '0';

    bypass_empty    <= '1';

    case state is
        when S_RESET =>
            c_piso_next	   <= (others => '0');
            c_sipo_next	   <= (others => '0');
            c_Wcoef_next   <= (others => '0');
            c_Rcoef_next   <= (others => '0');
            c_hashw8_next  <= (others => '0');
            c_hashw32_next <= (others => '0');
            wait_next	   <= '0';
            sipo_full_next <= '0';
            bypass_empty_next <= '1';
            done_next       <= '0';
            chk_failed_next <= '0';
            hash_done_next  <= '0';
            state_next     <= S_FIRST_LOAD;

        when S_FIRST_LOAD =>
            if (infifo_empty = '0') then
                infifo_read <= '1';
                en_piso     <= '1';
                ld_piso     <= '1';
                c_hashw32_next <= c_hashw32_r + 1;
                state_next  <= S_SHIFT_LOAD;
            end if;

        when S_SHIFT_LOAD =>
            if (hash_done_r = '0') then
                if (fifo_full = '0') then
                    if (c_piso_r < 3) or ((c_piso_r = 3 )and(c_hashw32_r = NUM_HASH32_REQ_MGF))then
                        en_piso       <= '1';
                        c_piso_next   <= c_piso_r + 1;
                        c_hashw8_next <= c_hashw8_r + 1;
                        if (lseq_243 = '1') then
                            fifo_write  <= '1';
                            c_Wcoef_next <= c_Wcoef_r + 1;
                        end if;
                        if (c_piso_r = 3 )and(c_hashw32_r = NUM_HASH32_REQ_MGF) then
                            hash_done_next <= '1';
                        end if;
                    else
                        if (infifo_empty = '0') then
                            c_piso_next	 <= (others => '0');
                            infifo_read <= '1';
                            en_piso     <= '1';
                            ld_piso     <= '1';
                            c_hashw32_next <= c_hashw32_r + 1;
                            if (wait_r = '0') then
                                c_hashw8_next <= c_hashw8_r + 1;
                                if (lseq_243 = '1') then
                                    fifo_write  <= '1';
                                    c_Wcoef_next <= c_Wcoef_r + 1;
                                end if;
                            else
                                wait_next <= '0';
                            end if;
                        elsif (wait_r = '0') then
                            wait_next <= '1';
                            c_hashw8_next <= c_hashw8_r + 1;
                            if (lseq_243 = '1') then
                                fifo_write  <= '1';
                                c_Wcoef_next <= c_Wcoef_r + 1;
                            end if;
                        end if;
                    end if;
                end if;
            end if;

            if (sipo_full_r = '0') then
                if (fifo_empty = '0') then
                    fifo_read    <= '1';
                    c_Rcoef_next <= c_Rcoef_r + 1;
                    en_sipo      <= '1';
                    if (c_sipo_r < 63) then
                        c_sipo_next <= c_sipo_r + 1;
                    else
                        c_sipo_next	<= (others => '0');
                        sipo_full_next   <= '1';
                    end if;
                end if;
            else
                sel_o         <= c_sipo_r(2 downto 0);
                output_valid  <= '1';
                if (output_ready = '1') then
                    if (c_sipo_r < 4) then
                        c_sipo_next <= c_sipo_r + 1;
                    else
                        if (done_r = '1') or (chk_failed_r = '1') then
                            state_next  <= S_WAIT;
                        end if;
                        c_sipo_next	<= (others => '0');
                        sipo_full_next <= '0';
                    end if;
                end if;
            end if;

            case checkpoint is
                when "000" =>
                    if (c_hashw8_r >= 96) then
                        if (c_Wcoef_r < 64)  then
                            chk_failed_next <= '1';
                        else
                            checkpoint_next   <= "001";
                        end if;
                    end if;

                when "001" =>
                    if (c_hashw8_r >= 160) then
                        if (c_Wcoef_r < 128) then
                            chk_failed_next <= '1';
                        else
                            checkpoint_next   <= "010";
                        end if;
                    end if;

                    if (c_Rcoef_r > 63) then
                        bypass_empty <= '1';
                    else
                        bypass_empty <= '0';
                    end if;

                when "010" =>
                    if (c_hashw8_r >= 224) then
                        if (c_Wcoef_r < 192) then
                            chk_failed_next <= '1';
                        else
                            checkpoint_next   <= "011";
                        end if;
                    end if;

                    if (c_Rcoef_r > 127) then
                        bypass_empty <= '1';
                    else
                        bypass_empty <= '0';
                    end if;

                when "011" =>
                    if (c_hashw8_r >= 288) then
                        if (c_Wcoef_r < 256) then
                            chk_failed_next <= '1';
                        else
                            if (PARAM_N = 1087) then
                                checkpoint_next   <= "101";
                            else
                                checkpoint_next   <= "100";
                            end if;
                        end if;
                    elsif (PARAM_N = 1087) and (hash_done_r = '1')then
                        chk_failed_next <= '1';
                    end if;

                    if (c_Rcoef_r > 191) then
                        bypass_empty <= '1';
                    else
                        bypass_empty <= '0';
                    end if;

                when "100" =>
                    if (c_hashw8_r >= 352) then
                        if (c_Wcoef_r < 320) then
                            chk_failed_next <= '1';
                        else
                            checkpoint_next   <= "101";
                        end if;
                    elsif (hash_done_r = '1')then
                        chk_failed_next <= '1';
                    end if;

                    if (c_Rcoef_r > 255) then
                        bypass_empty <= '1';
                    else
                        bypass_empty <= '0';
                    end if;

                when others =>
                    if (PARAM_N = 1087) then
                        if (c_Rcoef_r > 255) then
                            done_next    <= '1';
                            bypass_empty <= '1';
                        else
                            bypass_empty <= '0';
                        end if;
                    else
                        if (c_Rcoef_r > 319) then
                            done_next    <= '1';
                            bypass_empty <= '1';
                        else
                            bypass_empty <= '0';
                        end if;
                    end if;
            end case;

        when S_WAIT =>
            bypass_empty <= '1';

    end case;
end process;

--! DATAPATH
Param_data_size_1087: if PARAM_N = 1087 generate
NUM_HASH32_REQ_MGF <= NUM_HASH_REQ_BWC2_1087*8;
end generate;

Param_data_size_1499: if PARAM_N = 1499 generate
NUM_HASH32_REQ_MGF <= NUM_HASH_REQ_BWC2_1499*8;
end generate;

done       <= done_r;
chk_failed <= chk_failed_r;

Piso: entity work.piso
    generic map(
        N => W,
        M => piso_width
        )
    port map(
        clk    => clk,
        rst    => rst,
        en     => en_piso,
        ld     => ld_piso,
        input  => infifo_dout,
        output => piso_out
    );

lseq_243 <= '1' when (piso_out < x"F3") else '0';

o2t_rom: entity work.o2t_rom(arch)
    port map(
        addr   => piso_out,
        output => o2t_out
    );

FIFO_FWFT: entity work.fifo(structure)
    generic map(
        G_LOG2DEPTH    => 7,
        G_W            => 10
    )
    port map (
        clk              => clk,
        rstn             => not(rst),
        full             => fifo_full,
        write            => fifo_write,
        di_data          => o2t_out,
        empty            => fifo_empty_orig,
        read             => fifo_read,
        do_data          => fifo_dout,
        almost_full      => open,
        almost_empty     => open
    );

fifo_empty <= '1' when (bypass_empty = '1') else fifo_empty_orig;

Sipo: entity work.sipo
    generic map(
        N => 64*10,
        M => 10
        )
    port map(
        clk    => clk,
        rst    => rst,
        en     => en_sipo,
        input  => fifo_dout,
        output => r640
    );

with sel_o select
    output_data <= r640(639 downto 512) when "000",
    		       r640(511 downto 384) when "001",
    		       r640(383 downto 256) when "010",
    		       r640(255 downto 128) when "011",
    		       r640(127 downto 0  ) when "100",
    		       (others => '0') when others;

end arch;
