--=========================================================
-- Project      : Hardware Implementation of NTRU Cryptosystems
-- Author       : Kevin Briggs
--=========================================================
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;

-- parallel in with serial and parallel out options
entity piso_w_PO_SI is
	generic (
	    IP_WIDTH : integer := 40;
	    OS_WIDTH : integer := 12;
	    OP_WIDTH : integer := 40
	);
	port
	(
		clk 	: in std_logic;
		rst     : in std_logic;
		en   	: in std_logic;
		ld	 	: in std_logic;
		pin  	: in std_logic_vector(IP_WIDTH-1 downto 0);
        sin  	: in std_logic_vector(OS_WIDTH-1 downto 0);
		pout 	: out std_logic_vector(OP_WIDTH-1 downto 0);
		sout 	: out std_logic_vector(OS_WIDTH-1 downto 0)
	);
end piso_w_PO_SI;

architecture arch_full of piso_w_PO_SI is

     -- constants
     constant NUM_WORDS      : natural := IP_WIDTH / OS_WIDTH;
     constant RMNDR          : natural := IP_WIDTH - (NUM_WORDS * OS_WIDTH);
     constant FILL_WIDTH     : natural := (OS_WIDTH - RMNDR);
     constant EXTRA          : integer := (RMNDR + OS_WIDTH - 1) / OS_WIDTH;

 	-- signals
 	constant ZEROS      : std_logic_vector(FILL_WIDTH-1 downto 0) := (others => '0');
 	constant ZEROS_FULL : std_logic_vector(OS_WIDTH -1 downto 0) := (others => '0');

 	-- arrays
 	type word_array is array ( 0 to NUM_WORDS - 1 + EXTRA) of std_logic_vector(OS_WIDTH-1 downto 0);
 	type word_array_l is array (0 to NUM_WORDS + EXTRA) of std_logic_vector(OS_WIDTH-1 downto 0);
 	signal asig, reg_in : word_array;
 	signal reg_out : word_array_l;

 begin
     gen_main : for i in 0 to NUM_WORDS generate

         -- intermediate signals
         gen_intermediate : if(i < NUM_WORDS) generate
             asig(i) <= pin((NUM_WORDS-i+1)*(OS_WIDTH) - FILL_WIDTH - 1 downto (NUM_WORDS-i)*(OS_WIDTH) - FILL_WIDTH);
             pout((NUM_WORDS-i+1)*(OS_WIDTH) - FILL_WIDTH - 1 downto (NUM_WORDS-i)*(OS_WIDTH) - FILL_WIDTH) <= reg_out(i);
         end generate;

         reg_in(i) <= asig(i) when (ld = '1') else reg_out(i+1);
         --reg_in(i) <= asig(i) when (ld = '1') else sin;

         -------- registers -------
         gen_main_reg : entity work.regn_v2 (pos_edge) generic map (N => OS_WIDTH, init => ZEROS_FULL)
             port map(clk => clk, rst => rst, en => en, input => reg_in(i), output => reg_out(i));
     end generate;

     asig(NUM_WORDS) <= pin(RMNDR - 1 downto 0) & ZEROS;
     pout(RMNDR - 1 downto 0) <= reg_out(NUM_WORDS)(OS_WIDTH - 1 downto OS_WIDTH - RMNDR);
     -- reg_out(NUM_WORDS+1) <= reg_out(0); --(others => '0');
     reg_out(NUM_WORDS+1) <= sin; --(others => '0');
     sout <= reg_out(0);

 end arch_full;

-- -- constants
-- constant NUM_WORDS      : natural := IP_WIDTH / OS_WIDTH;
-- constant RMNDR          : natural := IP_WIDTH - (NUM_WORDS * OS_WIDTH);
-- constant FILL_WIDTH     : natural := (OS_WIDTH - RMNDR);
-- constant EXTRA          : integer := (RMNDR + OS_WIDTH - 1) / OS_WIDTH;
--
-- -- signals
-- constant ZEROS      : std_logic_vector(FILL_WIDTH-1 downto 0) := (others => '0');
-- constant ZEROS_FULL : std_logic_vector(OS_WIDTH -1 downto 0) := (others => '0');
--
-- -- arrays
-- type word_array is array ( 0 to NUM_WORDS - 1 + EXTRA) of std_logic_vector(OS_WIDTH-1 downto 0);
-- type word_array_l is array (0 to NUM_WORDS + EXTRA) of std_logic_vector(OS_WIDTH-1 downto 0);
-- signal asig, reg_in : word_array;
-- signal reg_out : word_array_l;
--
-- begin
-- gen_main : for i in 0 to NUM_WORDS generate
--
--     -- intermediate signals
--     gen_intermediate : if(i > 0) generate
--         asig(i) <= pin((i+1)*(OS_WIDTH) - FILL_WIDTH - 1 downto (i)*(OS_WIDTH) - FILL_WIDTH);
--         pout((i+1)*(OS_WIDTH) - FILL_WIDTH - 1 downto i*(OS_WIDTH) - FILL_WIDTH) <= reg_out(i);
--     end generate;
--
--     reg_in(i) <= asig(i) when (ld = '1') else reg_out(i+1);
--
--     -------- registers -------
--     gen_main_reg : entity work.regn_v2 (pos_edge) generic map (N => OS_WIDTH, init => ZEROS_FULL)
--         port map(clk => clk, rst => rst, en => en, input => reg_in(i), output => reg_out(i));
-- end generate;
--
-- asig(0) <= pin(RMNDR - 1 downto 0) & ZEROS;
-- pout(RMNDR - 1 downto 0) <= reg_out(0)(OS_WIDTH - 1 downto OS_WIDTH - RMNDR);
-- reg_out(NUM_WORDS+1) <= (others => '0');
-- sout <= reg_out(0);
--
-- end arch_full;
-- -- ==================================================================================================
-- architecture arch_2 of piso_w_PO is
-- constant regamount : integer := IP_WIDTH/OS_WIDTH;
-- type reg_array is array ( 0 to regamount-1 ) of std_logic_vector(OS_WIDTH-1 downto 0);
-- signal reg, mux : reg_array;
-- begin
-- mux(regamount-1) <= pin(OS_WIDTH-1 downto 0) when ld ='1' else reg(0);
-- mux_gen : for i in 0 to regamount-2 generate
--     mux(i) <= pin(IP_WIDTH-OS_WIDTH*i-1 downto IP_WIDTH-OS_WIDTH*i-OS_WIDTH) when ld = '1' else reg(i+1);
-- end generate;
--
-- regX_gen : for i in 0 to regamount-1 generate
--     regX : process ( clk )
--     begin
--         if rising_edge(clk) then
--             if ( en = '1' ) then
--                 reg(i) <= mux(i);
--             end if;
--         end if;
--     end process;
-- end generate;
--
-- sout <= reg(0);
-- p_output_gen: for i in 0 to regamount-1 generate
--    pout(IP_WIDTH-OS_WIDTH*i-1 downto IP_WIDTH-OS_WIDTH*i-OS_WIDTH) <= reg(i);
-- end generate;


-- end arch_2;
-- ==================================================================================================
-- ==================================================================================================
