-- =============================================================================
-- Copyright © 2017-2018 by Cryptographic Engineering Research Group (CERG),
-- ECE Department, George Mason University
-- Fairfax, VA, U.S.A.
-- Author: Kevin Briggs, Farnoud Farahmand
-- =============================================================================
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;

library work;
use work.sha2_pkg.all;
use work.sha3_pkg.all;
use work.ntru_pkg.all;

entity poly_mult is
generic
(
	--PIPE_N  : std_logic_vector(BETA-1 downto 0) := (others => '0');
	PIPE_N   : std_logic_vector;
	NUM      : integer;
	NUMWIDTH : integer;
	ROTWIDTH : integer;
	WORDSIZE : integer
);
port
(
	-- input
	clk 		: in std_logic;
	rst         : in std_logic;
    -- control inputs
    ce_piso_en  : in std_logic;
    ce_piso_ld  : in std_logic;
    he_sipo_en  : in std_logic;
    he_sipo_ld  : in std_logic;
    Rfe_piso_en : in std_logic;
    Rfe_piso_ld : in std_logic;
    sel_d       : in std_logic;
    sel_r       : in std_logic;
    sel_t        : in std_logic;
    c0             : in std_logic;
    -- data inputs
    r_F			: in std_logic_vector(ROTWIDTH-1 downto 0);
    h_e            : in std_logic_vector(NUMWIDTH*(WORDSIZE/NUMWIDTH)-1 downto 0);
    -- data outputs
	ce          : out std_logic_vector(2*NUMWIDTH*WORDSIZE-1 downto 0);
	R_fe		: out std_logic_vector(2*NUMWIDTH*WORDSIZE-1 downto 0);
    cR          : in std_logic_vector(2*NUMWIDTH*WORDSIZE-1 downto 0);
    en_pipe     : in std_logic
);
end poly_mult;

architecture arch_full of poly_mult is

	--arrays
	type x_array is array(0 to NUM-1) of std_logic_vector(NUMWIDTH-1 downto 0);
	type ro_i_array is array(0 to NUM-1) of std_logic_vector(NUMWIDTH-1 downto 0);
	type sipo_out_array is array(0 to NUM) of std_logic_vector(NUMWIDTH-1 downto 0);

	signal sum_i         	: x_array;
	signal sum_fb_i, t, u, sum2_fb_i   : x_array;
	signal ro_i	 			: ro_i_array;
	signal rxi              : x_array;


	-- signals
	signal sum, sum_fb		: std_logic_vector(NUMWIDTH*NUM-1 downto 0);
	signal ro				: std_logic_vector(NUMWIDTH*NUM-1 downto 0) := (others => '0');
	signal sipo_out, sipo_out_rev   : std_logic_vector(NUMWIDTH*NUM-1 downto 0);
	signal piso0_out, piso0_out_rev : std_logic_vector(NUMWIDTH*NUM-1 downto 0);
	signal c0v              : std_logic_vector(NUMWIDTH-1 downto 0);
	signal r_f_z            : std_logic_vector(ROTWIDTH-1 downto 0);

    --!!!!!!!! JUST FOR TEST !!!!!!!!!
    -- signal sum_fb_test      : std_logic_vector(NUMWIDTH*NUM-1 downto 0);
    -- signal R_fe_11_test     : std_logic_vector(10 downto 0);

    --DEBUG
    -- type debug_array is array (0 to 1086) of std_logic_vector(10 downto 0);
    -- signal sipo_out_debug : debug_array;
    --DEBUG

begin
    --DEBUG
    -- GEN_DEBUG0: for i in 0 to 1086 generate
    --     sipo_out_debug(i) <= sipo_out(11*(i+1)-1 downto 11*i);
    -- end generate;
    --DEBUG

    GEN_Endian_Reverse: for i in 0 to NUM-1 generate
        sipo_out_rev(11*(i+1)-1 downto 11*i)  <= sipo_out(11*(NUM-i)-1 downto 11*(NUM-i)-11);
        piso0_out_rev(11*(i+1)-1 downto 11*i) <= piso0_out(11*(NUM-i)-1 downto 11*(NUM-i)-11);
    end generate;

    c0v <= (others => c0);

    r_f_z <= r_F when sel_r = '0' else
             (others => '0');

    -- Input SIPO
	sipo_w_pi: entity work.sipo_w_PI (lsbf)
	    generic map ( IS_WIDTH => NUMWIDTH*(WORDSIZE/NUMWIDTH), IP_WIDTH => NUM*NUMWIDTH, OP_WIDTH => NUM*NUMWIDTH )
	    port map ( clk => clk, rst => rst, en => he_sipo_en, ld => he_sipo_ld, sin => h_e, pin => piso0_out_rev, pout => sipo_out);
	-- Input Storage PISO
    piso_w_po_si_0 : entity work.piso_w_PO_SI (arch_full) generic map (IP_WIDTH => NUMWIDTH*NUM, OS_WIDTH => 2*NUMWIDTH*WORDSIZE, OP_WIDTH => NUMWIDTH*NUM)
        port map ( clk => clk, rst => rst, ld => ce_piso_ld, en => ce_piso_en, pin => sipo_out_rev, pout => piso0_out, sin => cR, sout => ce);
        --sipo_out

    -- Variable Rotator
	vrl : entity work.vrl (arch) generic map (PIPE_N => PIPE_N, NUM => NUM, NUMWIDTH => NUMWIDTH, ROTWIDTH => ROTWIDTH)
	    port map (clk => clk, rst => rst, A => sipo_out, B => r_f_z, C => ro, en_pipe => en_pipe);

	-- Coefficient Signals
	ii : for i in 0 to NUM-1 generate
		ro_i(i) <= ro(NUMWIDTH*(i+1)-1 downto NUMWIDTH*i);
	end generate;

	-- Combinational Blocks
	ii2 : for i in 0 to NUM-1 generate
        sum_fb_i(NUM-1-i) 					    <= sum_fb(NUMWIDTH*(i+1)-1 downto NUMWIDTH*i);
		sum2_fb_i(i)                        <= sum_fb_i(i)(NUMWIDTH-2 downto 0) & '0';
		t(i)								<= sum_fb_i(i) when (sel_t = '1') else (others => '0');
		rxi(i)                              <= ro_i(i) xor c0v;
		u(i)                                <= sum2_fb_i(i) when sel_d = '1' else rxi(i);
		sum_i(i) 							<= std_logic_vector(unsigned(u(i)) + unsigned(t(i)) + unsigned'(""&c0));
        sum(NUMWIDTH*(i+1)-1 downto NUMWIDTH*i)	<= sum_i(NUM-1-i);
	end generate;

	-- Output PISO
	piso_w_po_1 : entity work.piso_w_PO (arch_full) generic map (IP_WIDTH => NUMWIDTH*NUM, OS_WIDTH => 2*NUMWIDTH*WORDSIZE, OP_WIDTH => NUMWIDTH*NUM)
		port map ( clk => clk, rst => rst, ld => Rfe_piso_ld, en => Rfe_piso_en, pin => sum, pout => sum_fb, sout => R_fe);

    --DEBUG
--    piso_w_po_tb : entity work.piso (msb_first) generic map (N => NUMWIDTH*NUM, M => 11)
--		port map ( clk => clk, rst => rst, ld => Rfe_piso_ld, en => Rfe_piso_en, input => sum, --output => R_fe_11_test);


end arch_full;
