-- =====================================================================
-- Copyright © 2010-2012 by Cryptographic Engineering Research Group (CERG),
-- ECE Department, George Mason University
-- Fairfax, VA, U.S.A.
-- Author: CERG and Farnoud Farahmand
-- =====================================================================

library ieee ;
use ieee.std_logic_1164.all;
use work.sha3_pkg.all;
use work.sha2_pkg.all;

entity sha2_control_rs is
generic (fifo_mode	:integer :=ZERO_WAIT_STATE);
port(
	clk					:in std_logic;
	rst						:in std_logic;

	-- from datpath
	z16						:in std_logic;
	zlast					:in std_logic;
	o8						:in std_logic;
	last_block				:in std_logic;
	msg_done				:in std_logic;

	-- to datpath
	sel2					:out std_logic;
	sel						:out std_logic;
	sel_gh					:out std_logic;
	sel_gh2					:out std_logic;
	ctrl_rst				:out std_logic;
	wr_chctr				:out std_logic;
	rst_flags				:out std_logic;
	wr_data					:out std_logic;
	kw_wr					:out std_logic;
	wr_state				:out std_logic;
	wr_result				:out std_logic;
	wr_len					:out std_logic;
	ctr_ena					:out std_logic;

	-- fifo
	src_read				:out std_logic;
	src_ready				:in std_logic;
	dst_ready				:in std_logic;
	dst_write				:out std_logic;


	-- pad
    comp_rem_e0, comp_lb_e0 : in std_logic;
    en_lb, clr_lb : out std_logic;
    spos, sel_pad : out std_logic_vector(1 downto 0);

    --! NTRU control signal
    backup_ready      : out std_logic;
    sel_input_state   : out std_logic
);
end sha2_control_rs;



architecture zws of sha2_control_rs is


    type state_type is (idle,
						get_len,
						get_data,
						wr_get_len,
						wr_get_data,
						kw_compute,
						data,
						data_end,
						dummy,
						computation,
						computation_end,
						result_delay,
						result_delay2,
						result_delay3,
						result,
						mux_gh,
						send_data,
						chunk_done,
						done);

signal next_state, current_state: state_type;
attribute enum_encoding : string;


signal set_start_pad, clr_start_pad, start_pad : std_logic;
signal set_more_block, clr_more_block, more_block : std_logic;
signal set_valid_block, clr_valid_block, valid_block : std_logic; -- set that there are more block following this one
signal src_read_sig : std_logic;

begin

    state_reg: process(clk, rst)
    begin

	if (rst='1') then
            current_state <= done; -- set to done to reset all the flags
	elsif (clk'event and clk='1') then
	    current_state <= next_state;
	end if;

    end process;

    comb_logic: process(current_state,src_ready, z16, zlast,o8,last_block, msg_done, next_state, dst_ready, comp_rem_e0, more_block)
    begin
	case current_state is

	    when idle =>
			if src_ready='1' then
			    next_state <= get_len;
			else
			    next_state <= idle;
			end if;

	    when get_len =>

				if src_ready='1' then
			    	next_state <= get_data;
				else
			    	next_state <= get_len;
				end if;
		when get_data =>
			next_state <= wr_get_len;

	    when wr_get_len =>
			if src_ready='1' then
			    next_state <= wr_get_data;
			else
			    next_state <= wr_get_len;
			end if;

	    when wr_get_data =>
			if src_ready='1' or (last_block = '1' and msg_done = '1' and comp_rem_e0 = '1') then
			    next_state <= kw_compute;
			else
			    next_state <= wr_get_data;
			end if;

		when kw_compute =>
			if src_ready='1' or (last_block = '1' and msg_done = '1' and comp_rem_e0 = '1') then
			    next_state <= data;
			else
			    next_state <= kw_compute;
			end if;

	    when data =>
			if (src_ready='1' and z16 = '1') or (last_block = '1' and msg_done = '1' and comp_rem_e0 = '1' and z16 = '1') then
                next_state <= data_end;
           	else
            	next_state <= data;
			end if;

	    when data_end =>
			if src_ready='1' or (last_block = '1' and msg_done = '1' and comp_rem_e0 = '1') then
	 	   		next_state <= dummy;
			else
			    next_state <= data_end;
			end if;

		when dummy =>
			if src_ready='1' or (last_block = '1' and msg_done = '1' and comp_rem_e0 = '1') then
	 	   		next_state <= computation;
			else
			    next_state <= dummy;
			end if;


	    when computation =>
			if zlast='1' then
			    next_state <= computation_end;
			else
			    next_state <= computation;
			end if;

		when computation_end=>

			next_state <= result_delay;


		when result_delay=>
				next_state <= result_delay2;

		when result_delay2=>
				next_state <= result_delay3;

		when result_delay3=>
			if src_ready='1' or (last_block = '1' and msg_done = '1' and comp_rem_e0 = '1') then
				next_state <= result;
			else
				next_state <= result_delay3;
			end if;
	    when result =>
			if last_block='1' and msg_done='1' and start_pad = '0' and more_block = '0' then
				next_state <= send_data;
			elsif last_block='0' and msg_done='1'then
				next_state <= chunk_done;
			--elsif src_ready='1' or (msg_done = '1' and last_block = '1' and start_pad = '1') or more_block = '1' then
			elsif src_ready='1' or (msg_done = '1' and last_block = '1' and comp_rem_e0 = '1') then
				next_state <= mux_gh;
			else
				next_state <= result;
			end if;

		when mux_gh =>
			if src_ready='1' or (msg_done = '1' and last_block = '1' and comp_rem_e0 = '1') then
				next_state <= data;
			else
				next_state <= mux_gh;
			end if;

	    when send_data =>

			if dst_ready='0' and o8 = '1' then
		    	next_state <= done;
			else
		    	next_state <= send_data;
			end if;

		when done =>
			next_state <=get_len;

		when chunk_done =>
			next_state <=get_len;

	end case;

    end process;

	wr_len<= '1' when (current_state= get_len and src_ready = '1') else '0';

	wr_data <= '1' when
	((src_ready = '1' or (msg_done = '1' and comp_rem_e0 = '1')) and
	(
	(current_state=wr_get_len) or (current_state=wr_get_data) or (current_state=kw_compute) or (current_state=data) or (current_state=data_end)
	or (current_state=dummy)
	or (current_state=result)
	or (current_state=mux_gh)
	or ((current_state=result_delay3 or current_state=result) and (valid_block = '1' and (src_ready = '1' or comp_rem_e0 = '1')  and (last_block = '1' or (last_block = '0' and msg_done = '0'))))
	)) or (current_state=computation) or (current_state=computation_end) or (current_state=result_delay) or (current_state=result_delay2)
	else '0';

 	wr_result<= '1' when  (current_state= result_delay)
	or (current_state= result_delay2)
	or (current_state= result_delay3 and (src_ready='1' or (msg_done = '1' and last_block = '1' and comp_rem_e0 = '1')))
	or (current_state=result and (src_ready='1' or (msg_done = '1' and last_block = '1' and comp_rem_e0 = '1'))) else '0';



	src_read_sig <= '1' when (current_state=get_len and src_ready = '1')
	or
	((current_state=result_delay3 or current_state=result) and src_ready = '1'  and valid_block = '1' and (msg_done = '0' or (msg_done = '1' and comp_rem_e0 = '0' and start_pad = '1')))
	or
	(( current_state=wr_get_len
    or current_state=wr_get_data
	or current_state=kw_compute
    or current_state=data
    or current_state=data_end
    or current_state=dummy
    or current_state=mux_gh) and src_ready = '1'  and (last_block = '0' or msg_done = '0' or (start_pad = '1' and comp_rem_e0 = '0'))
	) else '0';
    src_read <= src_read_sig;

	wr_state <= '1' when
	((src_ready = '1' or (msg_done = '1' and comp_rem_e0 = '1')) and
	(
	(current_state=data_end or current_state=dummy)
	or (current_state = kw_compute)
	or (current_state = data)
	or (current_state=result)
	or (current_state=mux_gh)))
	or (current_state=result_delay3 and (src_ready='1' or (msg_done = '1' and last_block = '1' and comp_rem_e0 = '1')))
	or
	(current_state=computation or current_state=computation_end or current_state=result_delay or current_state=result_delay2)
    --! NTRU control
    or
    (current_state=done)
    else '0';

	ctr_ena <= '1' when
	((src_ready = '1' or (msg_done = '1' and comp_rem_e0 = '1'))
	and
	((current_state=wr_get_data)
	or (current_state=kw_compute)
	or (current_state=data)
	or (current_state=data_end)
	or (current_state=dummy)
	or (current_state=mux_gh)
	)) or
	(current_state=result and ((src_ready='1' and (msg_done='0' or (last_block='1' and comp_rem_e0='0'))) or (msg_done = '1' and last_block = '1' and comp_rem_e0 = '1')))
	or (current_state=computation) or (current_state=computation_end) or (current_state=result_delay) or (current_state=result_delay2)
	else '0';
	--(current_state=result and ((src_ready='1' and msg_done='0') or (last_block = '1' and valid_block = '1'))) else '0';

	kw_wr <='1' when
	((src_ready = '1' or (msg_done = '1' and comp_rem_e0 = '1')) and
	(
	(current_state=kw_compute)
	or (current_state = wr_get_data)
	or (current_state=data)
	or (current_state=data_end)
	or (current_state=dummy)
	or (current_state=result and (valid_block = '1' and (src_ready = '1' or comp_rem_e0 = '1') and (last_block = '1' or (last_block = '0' and msg_done = '0'))))
	or (current_state=mux_gh))) or (current_state=computation) or (current_state=computation_end) or (current_state=result_delay) or (current_state=result_delay2)
    --! NTRU control
    or
    (current_state=done)
	else '0';

	sel	<= '1' when (current_state=result) else '0';

	sel2 <='1' when  (current_state=computation) or (current_state=computation_end) or (current_state=result_delay)  else '0';

	sel_gh <= '1' when (current_state=wr_get_data) or (current_state=result)  else '0';

	sel_gh2 <= '1' when  (current_state=result) else '0';

	dst_write <= '1' when (current_state=send_data and dst_ready = '0') else '0';

	ctrl_rst <= '1' when (current_state=done) else '0';
    --! NTRU
    sel_input_state <= '1' when (current_state=done) else '0';
    backup_ready    <= '1' when (current_state=chunk_done) else '0';
    --! --------------------------------------------------------------

	rst_flags  <= '1' when (current_state=chunk_done) else '0';

	wr_chctr <= '1' when (current_state=dummy and msg_done = '0' and (src_ready='1' or (last_block = '1' and msg_done = '1' and comp_rem_e0 = '1')) )else '0';


    -- PADDING
    -- select pad control

	sel_pad(0) <= '1' when (current_state = dummy) else '0';
	sel_pad(1) <= '0' when ((current_state = data_end) or (current_state =  dummy))
                            and (last_block = '1' and msg_done = '1' and comp_lb_e0 = '1' and start_pad = '0' and more_block = '0') else '1';

    en_lb <= '1' when (src_read_sig = '1' and last_block = '1' and msg_done = '1' and comp_lb_e0 = '0') else '0';
    clr_lb <= '1' when (src_read_sig = '1' and last_block = '1' and msg_done = '1' and comp_lb_e0 = '1') else '0';

    sr_more_block :  sr_reg
        port map ( rst => rst, clk => clk, set => set_more_block, clr => clr_more_block, output => more_block);
    set_more_block <= '1' when
		-- for most case
		(current_state=data and z16 = '1' and last_block = '1' and msg_done = '1' and comp_rem_e0 = '0' and comp_lb_e0 = '0')   or
		-- for special case (msg = blocksize)
		(current_state=dummy and msg_done = '0')
		else '0';
	clr_more_block <= '1' when (current_state = mux_gh or current_state = chunk_done) else '0';


    sr_pos1 :  sr_reg
        port map ( rst => rst, clk => clk, set => set_start_pad, clr => clr_start_pad, output => start_pad);

	set_start_pad <= '1' when (current_state=get_data and last_block = '1' and msg_done = '1')
		or (current_state=dummy and msg_done = '0' and last_block = '1') else '0';

	clr_start_pad <= '1' when
		(current_state = wr_get_len or current_state = wr_get_data or current_state = kw_compute or current_state = data or current_state = data_end or current_state = dummy
		or current_state = result_delay3 or current_state = result or current_state = mux_gh)
		and
		(last_block = '1' and msg_done = '1' and comp_lb_e0 = '1' and start_pad = '1' and (src_ready = '1' or comp_rem_e0 = '1')) else '0';

    spos <= "00" when (last_block = '1' and msg_done = '1' and comp_lb_e0 = '1' and start_pad = '0') else	-- select '0'
			"01" when (last_block = '1' and msg_done = '1' and comp_lb_e0 = '1' and start_pad = '1') else 	-- select start pad
			"11";

	sr_valid_block :  sr_reg
        port map ( rst => rst, clk => clk, set => set_valid_block, clr => clr_valid_block, output => valid_block);
    set_valid_block <= '1' when ((current_state= computation_end or current_state=get_data ) and (msg_done = '0' or (msg_done = '1' and more_block = '1'))) else '0';
	clr_valid_block <= '1' when (current_state = dummy) else '0';
end zws;
