-- =============================================================================
-- Copyright © 2017-2018 by Cryptographic Engineering Research Group (CERG),
-- ECE Department, George Mason University
-- Fairfax, VA, U.S.A.
-- Author: Farnoud Farahmand
-- =============================================================================
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.std_logic_unsigned.all;
library work;
use work.ntru_pkg.all;

entity top_controller is
    port(
        clk             : in  std_logic;
        rst             : in  std_logic;

        pdi_data        : in std_logic_vector(PW-1 downto 0); -- we may remove it later
        pdi_valid       : in std_logic;
        pdi_ready       : out std_logic;

        sdi_data        : in std_logic_vector(SW-1 downto 0);
        sdi_valid       : in std_logic;
        sdi_ready       : out std_logic;

        rdi_valid       : in std_logic;
        rdi_ready       : out std_logic;

        ctrl_dout       : out std_logic_vector(31 downto 0);
        sel_ctrl_header : out std_logic;
        ctrl_do_valid   : out std_logic;
        ctrl_do_ready   : in std_logic;

        he_fifo_read    : in std_logic;
        he_fifo_empty   : out std_logic;

        ce_valid        : in std_logic;
        ce_read         : out std_logic;

        bpgm_mgf_start  : out std_logic;
        chk_failed_bwc1 : in std_logic;
        chk_failed_bwc2 : in std_logic;
        bpgm_mgf_done   : in std_logic;
        bpgm_mgf_sel    : out std_logic;
        bpgm_mgf_rst    : out std_logic;
        mask_ready      : out std_logic;
        mask_valid      : in std_logic;
        msg_len         : out std_logic_vector(7 downto 0);
        bpgm_mgf_Rferd  : in std_logic;

        decrypt         : out std_logic;
        sel_param       : out std_logic;
        check_3         : in std_logic;

        check_1         : in std_logic;
        check1_rst      : out std_logic;
        check1_en_reg   : out std_logic;
        check1_last     : out std_logic;

        poly_mult_rst   : out std_logic;
        poly_mult_update_h  : out std_logic;
        poly_mult_sdi_ready : out std_logic;
        poly_mult_start : out std_logic;
        poly_mult_done  : in std_logic;
        poly_mult_rF_sel: out std_logic;
        R_fe_valid_to_ctrl : in std_logic;
        Rfe_rd_from_ctrl   : out std_logic;
        sel_Rfe_rd         : out std_logic;

        sel_r_F         : out std_logic;
        sel_m_cm_di     : out std_logic;
        sel_b_cb_di     : out std_logic;
        sel_odcu        : out std_logic;
        sel_ci_mpr      : out std_logic;

        --! hTrunc Distributed RAM ports
        hTrunc_we       : out std_logic;
        hTrunc_addr_ctrl: out std_logic_vector(2 downto 0);
        sel_hTrunc_addr : out std_logic;
        --! F Distributed RAM ports
        sel_F_addr      : out std_logic;
        F_we            : out std_logic;
        F_addr_ctrl     : out std_logic_vector(7 downto 0);
        --! m_cm  BRAM ports
        m_cm_we         : out std_logic;
        m_cm_rd_ctrl    : out std_logic;
        m_cm_addr_ctrl  : out std_logic_vector(5 downto 0);
        sel_m_cm_addr   : out std_logic;
        --! b_cb Distributed RAM ports
        b_cb_we         : out std_logic;
        b_cb_addr_ctrl  : out std_logic_vector(2 downto 0);
        sel_b_cb_addr   : out std_logic;
        --! DFU-M controling signal
        dfum_sel_regoct : out std_logic;
        dfum_sel_data   : out std_logic;
        dfum_sel_sipo   : out std_logic;
        dfum_en_mreg    : out std_logic;
        dfum_en_sipo    : out std_logic;
        --! ODCU-E controling signal
        odcue_di_valid  : out std_logic;
        odcue_di_ready  : in std_logic;
        --! ODCU-D controling signal
        odcud_en_piso   : out std_logic;
        odcud_ld_piso   : out std_logic;
        odcud_en_mreg   : out std_logic;
        odcud_sel_cm    : out std_logic;
        odcud_sel_cb    : out std_logic;
        odcud_coctL     : in std_logic_vector(7 downto 0);
        odcud_check2    : in std_logic_vector(3 downto 0);
        --! IDCU-ED controling signal
        idcued_en_sipo_he   : out  std_logic;
        idcued_en_sipo_hram : out std_logic;
        idcued_en_piso_hram : out std_logic;
        idcued_ld_piso_hram : out std_logic;

        error : out std_logic

    );



end top_controller;

architecture behavioral of top_controller is

    --! STATE
    type t_state is (S_RESET, S_LD_PUBKEY_CT, S_INST_SEL, S_CHECK_HEADER,
        S_LD_PRIVKEY, S_ENCRYPT_1, S_ENCRYPT_2, S_ENCRYPT_3, S_ENCRYPT_4,
        S_ENCRYPT_5, S_ENCRYPT_6, S_ENCRYPT_7_A, S_ENCRYPT_7_B, S_ENCRYPT_8,
        S_UPDATE_b, S_LD_MSG_RAND, S_DECRYPT, S_CHECK_HEADER_SDI, S_DECRYPT_0,
        S_DECRYPT_2, S_DECRYPT_3, S_DECRYPT_3_WAIT, S_DECRYPT_4, S_DECRYPT_5,
        S_DECRYPT_5_WAIT, S_DECRYPT_6, S_DECRYPT_1, S_DECRYPT_7, S_DECRYPT_8,
        S_DECRYPT_9, S_DECRYPT_10, S_DECRYPT_11, S_DECRYPT_12, S_DECRYPT_13,
        S_DECRYPT_14, S_DECRYPT_14_WAIT, S_DECRYPT_15, S_LD_RAND);
    signal state, state_next : t_state;

    signal count_r, count_next, count_r_minus1   : std_logic_vector(10 downto 0);
    signal count2_r, count2_next                 : std_logic_vector(4 downto 0);
    signal count3_r, count3_next                 : std_logic_vector(4 downto 0);
    signal decrypt_r, decrypt_next               : std_logic;
    signal msg_len_r, msg_len_next               : std_logic_vector(7 downto 0);
    signal pubkey_loaded_r, pubkey_loaded_next   : std_logic;
    signal privkey_loaded_r, privkey_loaded_next : std_logic;
    signal instruction_r, instruction_next       : std_logic_vector(3 downto 0);
    signal bpgm_mgf_sel_r, bpgm_mgf_sel_next     : std_logic;
    signal r_F_sel_r, r_F_sel_next               : std_logic;
    signal db_32, p0Len_32, MbinPartial_32, DF, DEC_MSG_LEN_32, MbinSize_32,
    DEC_MSG_LEN_32_CEIL : integer;

    constant NUM_704_Blk: integer :=  divceil((PARAM_N*11), 704);
begin

    msg_len <= msg_len_r;
    decrypt <= decrypt_r;
    bpgm_mgf_sel <= bpgm_mgf_sel_r;
    poly_mult_sdi_ready <= privkey_loaded_r;
    poly_mult_rF_sel <= r_F_sel_r;
    sel_r_F <= r_F_sel_r;

    --DEC_MSG_LEN_32 <= divceil((to_integer(unsigned(msg_len_r))),4);
    DEC_MSG_LEN_32 <= (to_integer(unsigned(msg_len_r)))/4;
    DEC_MSG_LEN_32_CEIL <= divceil((to_integer(unsigned(msg_len_r))),4);

    TOP_CTRL_1087_PARAM_GEN: if PARAM_N = 1087 generate
        db_32    <= db_32_1087;
        MbinPartial_32  <=
            divceil((db_Byte_1087+1+to_integer(unsigned(msg_len_r))),4);
        p0Len_32 <= MbinSize_32_1087 - MbinPartial_32;
        DF       <= DF_1087;
        MbinSize_32 <= MbinSize_32_1087;
    end generate;

    TOP_CTRL_1499_PARAM_GEN: if PARAM_N = 1499 generate
        db_32 <= db_32_1499;
        MbinPartial_32  <=
            divceil((db_Byte_1499+1+to_integer(unsigned(msg_len_r))),4);
        p0Len_32 <= MbinSize_32_1499 - MbinPartial_32;
        DF       <= DF_1499;
        MbinSize_32 <= MbinSize_32_1499;
    end generate;

    p_fsm: process(clk)
    begin
        if rising_edge(clk) then
            if (rst = '1') then
                state   <= S_RESET;
                r_F_sel_r <= '0';
            else
                state       <= state_next;
                r_F_sel_r   <= r_F_sel_next;
            end if;
            count_r          <= count_next;
            count2_r         <= count2_next;
            count3_r         <= count3_next;
            msg_len_r        <= msg_len_next;
            decrypt_r        <= decrypt_next;
            pubkey_loaded_r  <= pubkey_loaded_next;
            privkey_loaded_r <= privkey_loaded_next;
            instruction_r    <= instruction_next;
            bpgm_mgf_sel_r   <= bpgm_mgf_sel_next;

        end if;
    end process;

        p_comb: process(state, pdi_data, pdi_valid, sdi_valid, rdi_valid,
            R_fe_valid_to_ctrl, he_fifo_read, mask_valid, ce_valid,
            chk_failed_bwc1, chk_failed_bwc2, bpgm_mgf_done, check_3,
            poly_mult_done, count_r, count2_r, pubkey_loaded_r, odcud_coctL,
            privkey_loaded_r, instruction_r, msg_len_r, check_1, odcue_di_ready,
            DEC_MSG_LEN_32, bpgm_mgf_Rferd, count3_r, ctrl_do_ready)

    begin
        --! Default values
        state_next       <= state;
        count_next       <= count_r;
        count2_next      <= count2_r;
        count3_next      <= count3_r;
        msg_len_next     <= msg_len_r;
        decrypt_next     <= decrypt_r;
        pubkey_loaded_next  <= pubkey_loaded_r;
        privkey_loaded_next <= privkey_loaded_r;
        instruction_next    <= instruction_r;
        bpgm_mgf_sel_next   <= bpgm_mgf_sel_r;
        r_F_sel_next        <= r_F_sel_r;

        pdi_ready        <= '0';
        sdi_ready        <= '0';
        rdi_ready        <= '0';

        ctrl_dout        <= (others => '0');
        sel_ctrl_header  <= '0';
        ctrl_do_valid    <= '0';

        Rfe_rd_from_ctrl <= '0';
        sel_Rfe_rd       <= '0';
        he_fifo_empty    <= '1';
        mask_ready       <= '0';
        ce_read          <= '0';

        bpgm_mgf_start   <= '0';
        bpgm_mgf_rst     <= '0';

        sel_param        <= '0';

        poly_mult_rst    <= '0';
        poly_mult_update_h  <= '0';
        poly_mult_start     <= '0';

        sel_m_cm_di      <= '0';
        sel_b_cb_di      <= '0';
        sel_hTrunc_addr  <= '0';
        sel_odcu         <= '0';
        sel_ci_mpr       <= '0';
        sel_m_cm_addr    <= '0';
        sel_b_cb_addr    <= '0';

        check1_rst       <= '0';
        check1_en_reg    <= '0';
        check1_last      <= '0';

        hTrunc_we        <= '0';
        hTrunc_addr_ctrl <= (others => '0');

        sel_F_addr       <= '0';
        F_we             <= '0';
        F_addr_ctrl      <= (others => '0');

        m_cm_we          <= '0';
        m_cm_rd_ctrl     <= '0';
        m_cm_addr_ctrl   <= (others => '0');

        b_cb_we          <= '0';
        b_cb_addr_ctrl   <= (others => '0');

        dfum_sel_regoct  <= '0';
        dfum_sel_data    <= '0';
        dfum_sel_sipo    <= '0';
        dfum_en_mreg     <= '0';
        dfum_en_sipo     <= '0';

        odcue_di_valid    <= '0';

        odcud_en_piso    <= '0';
        odcud_ld_piso    <= '0';
        odcud_en_mreg    <= '0';
        odcud_sel_cm     <= '0';
        odcud_sel_cb     <= '0';

        idcued_en_sipo_he   <= '0';
        idcued_en_sipo_hram <= '0';
        idcued_en_piso_hram <= '0';
        idcued_ld_piso_hram <= '0';

        error  <= '0';

        case state is
            when S_RESET =>
                pdi_ready  <= '1';
                if (pdi_valid = '1') then
                    count_next	     <= (others => '0');
                    count2_next	     <= (others => '0');
                    count3_next	     <= (others => '0');
                    msg_len_next     <= (others => '0');
                    pubkey_loaded_next  <= '0';
                    privkey_loaded_next <= '0';
                    decrypt_next        <= '0';
                    bpgm_mgf_sel_next   <= '0';
                    bpgm_mgf_rst        <= '1';
                    check1_rst          <= '1';
                    if (pdi_data(PW-1 downto PW-3) = PUBLIC_KEY)then
                        pdi_ready          <= '1';
                        poly_mult_start    <= '1';
                        poly_mult_update_h <= '1';
                        instruction_next <= (others => '1');
                        state_next       <= S_LD_PUBKEY_CT;
                    elsif (pdi_data(PW-1 downto PW-4) = INST_LD_PRIVKEY)then
                        pdi_ready        <= '1';
                        instruction_next <= INST_LD_PRIVKEY;
                        state_next       <= S_CHECK_HEADER_SDI;
                    else
                        error <= '1';
                    end if;
                end if;

            when S_INST_SEL =>
                if (pdi_valid = '1') then
                    if (pdi_data(PW-1 downto PW-3) = PUBLIC_KEY) then
                        pdi_ready          <= '1';
                        poly_mult_start    <= '1';
                        poly_mult_update_h <= '1';
                        state_next         <= S_LD_PUBKEY_CT;
                    elsif (pdi_data(PW-1 downto PW-4) = INST_LD_PRIVKEY) then
                        pdi_ready        <= '1';
                        instruction_next <= INST_LD_PRIVKEY;
                        -- state_next       <= S_CHECK_HEADER;
                        state_next       <= S_CHECK_HEADER_SDI;
                    elsif (pdi_data(PW-1 downto PW-4) = INST_ENCRYPT) then
                        pdi_ready      <= '1';
                        if (pubkey_loaded_r = '1') then
                            check1_rst       <= '1';
                            instruction_next <= INST_ENCRYPT;
                            r_F_sel_next     <= '0';
                            state_next <= S_CHECK_HEADER;
                        else
                            error      <= '1';
                            state_next <= S_RESET;
                        end if;
                    elsif (pdi_data(PW-1 downto PW-4) = INST_DECRYPT) then
                        pdi_ready      <= '1';
                        if ((pubkey_loaded_r = '1')and(privkey_loaded_r = '1')) then
                            decrypt_next     <= '1';
                            check1_rst       <= '1';
                            instruction_next <= INST_DECRYPT;
                            r_F_sel_next     <= '1';
                            state_next       <= S_CHECK_HEADER;
                        else
                            error      <= '1';
                            state_next <= S_RESET;
                        end if;
                    else
                        error      <= '1';
                        state_next <= S_RESET;
                    end if;
                end if;

            when S_CHECK_HEADER =>
                if (pdi_valid = '1') then
                    if (instruction_r = INST_ENCRYPT) then
                        pdi_ready          <= '1';
                        if (pdi_data(PW-1 downto PW-3) = MESSAGE) then
                            msg_len_next   <= pdi_data(7 downto 0);
                            state_next     <= S_LD_MSG_RAND;
                        else
                            error      <= '1';
                            state_next <= S_RESET;
                        end if;
                    elsif (instruction_r = INST_DECRYPT) then
                        pdi_ready          <= '1';
                        if (pdi_data(PW-1 downto PW-3) = CIPHERTEXT) then
                            msg_len_next     <= pdi_data(7 downto 0);
                            poly_mult_start  <= '1';
                            bpgm_mgf_sel_next <= '1';
                            state_next        <= S_LD_PUBKEY_CT;
                        else
                            error      <= '1';
                            state_next <= S_RESET;
                        end if;
                    end if;
                end if;

            when S_CHECK_HEADER_SDI =>
                if (sdi_valid = '1') then
                    if (instruction_r = INST_LD_PRIVKEY) then
                        sdi_ready          <= '1';
                        if (sdi_data(SW-1 downto SW-3) = PRIVATE_KEY) then
                            state_next    <= S_LD_PRIVKEY;
                        else
                            error      <= '1';
                            state_next <= S_RESET;
                        end if;
                    else
                        error      <= '1';
                        state_next <= S_RESET;
                    end if;
                end if;

            when S_LD_PUBKEY_CT =>
                if (pdi_valid = '1') then
                    if (count_r(0) = '0') then
                        pdi_ready  <= '1';
                        idcued_en_sipo_he   <= '1';
                        count_next <= count_r + 1;
                    else
                        he_fifo_empty <= '0';
                        if (he_fifo_read = '1') then
                            if (count_r = PARAM_N) then
                                count_next <= (others => '0');
                                if (decrypt_r = '0') then
                                    pubkey_loaded_next <= '1';
                                    state_next <= S_INST_SEL;
                                else
                                    if (privkey_loaded_r = '1') then
                                        bpgm_mgf_sel_next <= '1';
                                        state_next <= S_DECRYPT_0;
                                    else
                                        state_next <= S_INST_SEL;
                                    end if;
                                end if;
                            else
                                pdi_ready  <= '1';
                                count_next <= count_r + 1;
                            end if;
                        end if;
                    end if;

                    if (decrypt_r = '0') then
                        if (count_r < 24) then
                            idcued_en_sipo_hram <= '1';
                        elsif (count_r = 24) then
                            idcued_en_piso_hram <= '1';
                            idcued_ld_piso_hram <= '1';
                        elsif (count_r > 24) and (count_r < 33) then
                        --! this step can be removed later. there is no need for the hram,
                        --! because the data can be read directly from the piso in idcued
                            count_r_minus1   <= count_r - 1;
                            sel_hTrunc_addr  <= '1';
                            hTrunc_we        <= '1';
                            hTrunc_addr_ctrl <= count_r_minus1(2 downto 0);
                            idcued_en_piso_hram <= '1';
                        end if;
                    end if;
                end if;

            when S_LD_PRIVKEY =>
                F_addr_ctrl <= count_r(7 downto 0);
                sel_F_addr  <= '1';
                sdi_ready   <= '1';
                if (sdi_valid = '1') then
                    F_we       <= '1';
                    if (count_r = (2*DF)-1) then
                        count_next <= (others => '0');
                        privkey_loaded_next <= '1';
                        state_next <= S_INST_SEL;
                    else
                        count_next <= count_r + 1;
                    end if;
                end if;

            when S_LD_MSG_RAND =>
                sel_m_cm_addr  <= '1';
                m_cm_addr_ctrl <= count_r(5 downto 0);
                if (pdi_valid = '1') and (count_r < divceil(to_integer(unsigned(msg_len_r)), 4)) then
                    pdi_ready    <= '1';
                    m_cm_rd_ctrl <= '1';
                    m_cm_we      <= '1';
                    count_next   <= count_r + 1;
                end if;

                sel_b_cb_addr  <= '1';
                b_cb_addr_ctrl <= count2_r(2 downto 0);
                if (rdi_valid = '1') and (count2_r < db_32) then
                    rdi_ready  <= '1';
                    b_cb_we    <= '1';
                    count2_next <= count2_r + 1;
                end if;

                if (count_r = divceil( to_integer(unsigned(msg_len_r)), 4)) and (count2_r = db_32) then
                    count_next  <= (others => '0');
                    count2_next <= (others => '0');
                    if (instruction_r = INST_ENCRYPT) then
                        state_next <= S_ENCRYPT_1;
                    else
                        state_next <= S_DECRYPT;
                    end if;
                end if;

            when S_ENCRYPT_1 =>
                --! Start generating r and computing R = h * r
                bpgm_mgf_start  <= '1';
                poly_mult_start <= '1';
                state_next      <= S_ENCRYPT_2;

            when S_ENCRYPT_2 =>
                --! Wait for bpgm mgf to generate r and poly mult to calculate R
                if (chk_failed_bwc1 = '1') then
                    bpgm_mgf_rst <= '1';
                    state_next   <= S_UPDATE_b;
                elsif (poly_mult_done = '1') and (bpgm_mgf_done = '1') then
                    bpgm_mgf_sel_next <= '1';
                    state_next   <= S_ENCRYPT_3;
                end if;

            when S_ENCRYPT_3 =>
                --! start generating mask and sendig out poly_mult data
                poly_mult_start <= '1';
                -- bpgm_mgf_sel    <= '1';
                bpgm_mgf_start  <= '1';
                state_next      <= S_ENCRYPT_4;

            when S_ENCRYPT_4 =>
                --! waiting for the mask to be generated, generating Mbin using DFU-M
                sel_b_cb_addr  <= '1';
                b_cb_addr_ctrl <= count_r(2 downto 0);

                if (chk_failed_bwc2 = '1') then
                    bpgm_mgf_rst <= '1';
                    state_next   <= S_UPDATE_b;
                elsif (count2_r = 3) then
                    if (mask_valid = '1') and (R_fe_valid_to_ctrl = '1') and
                        (odcue_di_ready = '1') then
                        odcue_di_valid   <= '1';
                        check1_en_reg    <= '1';
                        sel_Rfe_rd       <= '1';
                        mask_ready       <= '1';
                        Rfe_rd_from_ctrl <= '1';
                        count2_next      <= (others => '0');
                    end if;
                -- elsif (count_r < db_32) then
                else
                    dfum_sel_data <= '1';
                    dfum_en_sipo  <= '1';
                    dfum_sel_sipo <= '1';
                    count2_next   <= count2_r + 1;
                    if (count_r = db_32-1) then
                        count_next   <= (others => '0');
                        state_next   <= S_ENCRYPT_5;
                    else
                        count_next   <= count_r + 1;
                    end if;
                end if;

            when S_ENCRYPT_5 =>
                sel_m_cm_addr  <= '1';
                m_cm_rd_ctrl   <= '1';
                m_cm_addr_ctrl <= count_r(5 downto 0);
                if (chk_failed_bwc2 = '1') then
                    bpgm_mgf_rst <= '1';
                    state_next   <= S_UPDATE_b;
                elsif (count2_r = 3) then
                    if (mask_valid = '1') and (R_fe_valid_to_ctrl = '1') and
                        (odcue_di_ready = '1') then
                        odcue_di_valid   <= '1';
                        check1_en_reg    <= '1';
                        sel_Rfe_rd       <= '1';
                        mask_ready       <= '1';
                        Rfe_rd_from_ctrl <= '1';
                        count2_next      <= (others => '0');
                    end if;
                else
                    count_next     <= count_r + 1;
                    state_next     <= S_ENCRYPT_6;
                end if;

            when S_ENCRYPT_6 =>
                --! Load message
                sel_m_cm_addr  <= '1';
                m_cm_rd_ctrl   <= '1';
                m_cm_addr_ctrl <= count_r(5 downto 0);
                if (chk_failed_bwc2 = '1') then
                    bpgm_mgf_rst <= '1';
                    state_next   <= S_UPDATE_b;
                elsif (count2_r = 3) then
                    if (mask_valid = '1') and (R_fe_valid_to_ctrl = '1') and
                        (odcue_di_ready = '1') then
                        odcue_di_valid   <= '1';
                        check1_en_reg    <= '1';
                        sel_Rfe_rd       <= '1';
                        mask_ready       <= '1';
                        Rfe_rd_from_ctrl <= '1';
                        count2_next      <= (others => '0');
                        count_next       <= count_r + 1;
                    end if;
                else
                    -- if (count_r <= MbinPartial_32 - db_32) then
                    if (count_r = 1) then
                        dfum_sel_regoct <= '1';
                    end if;
                    dfum_en_sipo  <= '1';
                    dfum_sel_sipo <= '1';
                    dfum_en_mreg  <= '1';
                    count2_next   <= count2_r + 1;
                    if (count_r = MbinPartial_32 - db_32) then --
                        count_next <= (others => '0');
                        if (count2_r <= 1) then
                            state_next <= S_ENCRYPT_7_A;
                        else
                            state_next <= S_ENCRYPT_7_B;
                        end if;
                    elsif (count2_r /= 2) then
                        count_next <= count_r + 1;
                    end if;
                end if;

            when S_ENCRYPT_7_A =>
                --! Do the partial padding
                if (chk_failed_bwc2 = '1') then
                    bpgm_mgf_rst <= '1';
                    state_next   <= S_UPDATE_b;
                else
                    dfum_en_sipo  <= '1';
                    count2_next   <= count2_r + 1;
                    if (count2_r = 2) then
                        state_next <= S_ENCRYPT_7_B;
                    end if;
                end if;


            when S_ENCRYPT_7_B =>
                --! Do the padding
                if (chk_failed_bwc2 = '1') then
                    bpgm_mgf_rst <= '1';
                    state_next   <= S_UPDATE_b;
                elsif (count2_r = 3) then
                    if (mask_valid = '1') and (R_fe_valid_to_ctrl = '1') and
                        (odcue_di_ready = '1') then
                        odcue_di_valid   <= '1';
                        check1_en_reg    <= '1';
                        sel_Rfe_rd       <= '1';
                        mask_ready       <= '1';
                        Rfe_rd_from_ctrl <= '1';
                        count2_next      <= (others => '0');
                        if (count_r = p0Len_32) then
                            check1_last  <= '1';
                            count_next <= (others => '0');
                            state_next <= S_ENCRYPT_8;
                        end if;
                    end if;
                else
                    dfum_en_sipo  <= '1';
                    count2_next   <= count2_r + 1;
                    count_next <= count_r + 1;
                end if;

            when S_ENCRYPT_8 =>
                bpgm_mgf_sel_next <= '0';
                if (check_1 = '1') then
                    count_next	     <= (others => '0');
                    count2_next	     <= (others => '0');
                    msg_len_next     <= (others => '0');
                    bpgm_mgf_sel_next   <= '0';
                    bpgm_mgf_rst        <= '1';
                    check1_rst          <= '1';
                    state_next       <= S_INST_SEL;
                else
                    state_next <= S_UPDATE_b;
                end if;

            when S_UPDATE_b =>
                bpgm_mgf_rst  <= '1';
                poly_mult_rst <= '1';
                count_next	  <= (others => '0');
                count2_next	  <= (others => '0');
                state_next    <= S_LD_RAND;

            when S_LD_RAND =>
                sel_b_cb_addr  <= '1';
                b_cb_addr_ctrl <= count2_r(2 downto 0);
                if (rdi_valid = '1') then
                    rdi_ready  <= '1';
                    b_cb_we    <= '1';
                    if (count2_r = db_32-1) then
                        count2_next	<= (others => '0');
                        state_next  <= S_ENCRYPT_1;
                    else
                        count2_next <= count2_r + 1;
                    end if;
                end if;

            when S_DECRYPT_0 =>
                --! Wait for generating ci by computing ci = f * e
                if (poly_mult_done = '1') then
                    bpgm_mgf_start  <= '1';
                    poly_mult_start <= '1';
                    state_next   <= S_DECRYPT_1;
                end if;

            when S_DECRYPT_1 =>
                if (bpgm_mgf_Rferd = '1') and (R_fe_valid_to_ctrl = '1')then
                    check1_en_reg    <= '1';
                end if;
                if (poly_mult_done = '1') then
                    check1_last  <= '1';
                    state_next   <= S_DECRYPT_2;
                end if;

            when S_DECRYPT_2 =>
                --! First load into the ODCU_D piso
                if check_1 = '0' then
                    error       <= '1';
                    state_next  <= S_RESET;
                elsif (mask_valid = '1') and (R_fe_valid_to_ctrl = '1')then
                    odcud_en_piso    <= '1';
                    odcud_ld_piso    <= '1';
                    sel_Rfe_rd       <= '1';
                    mask_ready       <= '1';
                    Rfe_rd_from_ctrl <= '1';
                    state_next   <= S_DECRYPT_3;
                end if;


            when S_DECRYPT_3 =>
                --! Load cb into b_cb RAM
                sel_b_cb_di    <= '1';
                sel_b_cb_addr  <= '1';
                b_cb_addr_ctrl <= count_r(2 downto 0);

                if (chk_failed_bwc2 = '1') then
                    error       <= '1';
                    state_next  <= S_RESET;
                else
                    if (count2_r = 2) then
                        if (mask_valid = '1') and (R_fe_valid_to_ctrl = '1')then
                            odcud_en_piso    <= '1';
                            odcud_ld_piso    <= '1';
                            sel_Rfe_rd       <= '1';
                            mask_ready       <= '1';
                            Rfe_rd_from_ctrl <= '1';
                            count2_next      <= (others => '0');

                            if (count_r = db_32-1) then
                                count_next  <= (others => '0');
                                state_next  <= S_DECRYPT_4;
                            else
                                count_next  <= count_r + 1;
                            end if;
                        else
                            if (count_r = db_32-1) then
                                count_next  <= (others => '0');
                                state_next  <= S_DECRYPT_3_WAIT;
                            else
                                count_next  <= count_r + 1;
                                state_next  <= S_DECRYPT_3_WAIT;
                            end if;
                        end if;
                    else
                        odcud_en_piso <= '1';
                        count2_next   <= count2_r + 1;
                        if (count_r = db_32-1) then
                            count_next  <= (others => '0');
                            state_next  <= S_DECRYPT_4;
                        else
                            count_next  <= count_r + 1;
                        end if;
                    end if;

                    b_cb_we  <= '1';

                end if;

            when S_DECRYPT_3_WAIT =>
                if (mask_valid = '1') and (R_fe_valid_to_ctrl = '1') then
                    odcud_en_piso    <= '1';
                    odcud_ld_piso    <= '1';
                    sel_Rfe_rd       <= '1';
                    mask_ready       <= '1';
                    Rfe_rd_from_ctrl <= '1';
                    count2_next      <= (others => '0');
                    if (count_r = 0) then
                        state_next <= S_DECRYPT_4;
                    else
                        state_next <= S_DECRYPT_3;
                    end if;
                end if;

            when S_DECRYPT_4 =>
                --! Get decrypted message length
                if (count2_r = 0) then
                    odcud_en_mreg <= '1';
                    msg_len_next  <= odcud_coctL;

                    odcud_en_piso <= '1';
                    count2_next   <= count2_r + 1;
                    state_next    <= S_DECRYPT_5;
                else
                    if (mask_valid = '1') and (R_fe_valid_to_ctrl = '1') then
                        odcud_en_mreg <= '1';
                        msg_len_next  <= odcud_coctL;

                        odcud_en_piso    <= '1';
                        odcud_ld_piso    <= '1';
                        sel_Rfe_rd       <= '1';
                        mask_ready       <= '1';
                        Rfe_rd_from_ctrl <= '1';
                        count2_next      <= (others => '0');
                        count3_next      <= count3_r + 1;
                        state_next <= S_DECRYPT_5;
                    end if;
                end if;

            when S_DECRYPT_5 =>
                --! Load cm into m_cm RAM
                sel_m_cm_di    <= '1';
                sel_m_cm_addr  <= '1';
                m_cm_addr_ctrl <= count_r(5 downto 0);

                if (chk_failed_bwc2 = '1') then
                    error       <= '1';
                    state_next  <= S_RESET;
                else
                    m_cm_rd_ctrl <= '1';
                    m_cm_we      <= '1';
                    if (count2_r = 2) then
                        if (count_r >= DEC_MSG_LEN_32-1) then
                            odcud_en_mreg <= '1';
                            if msg_len_r(1 downto 0) = "00" then
                                count_next <= (others => '0');
                                state_next <= S_DECRYPT_6;
                            else
                                if (count_r = DEC_MSG_LEN_32) then
                                    odcud_sel_cm <= '1';
                                    count_next <= (others => '0');
                                    state_next <= S_DECRYPT_6;
                                else
                                    count_next  <= count_r + 1;
                                end if;
                            end if;
                        elsif (mask_valid = '1') and (R_fe_valid_to_ctrl = '1')then
                            odcud_en_mreg    <= '1';
                            odcud_en_piso    <= '1';
                            odcud_ld_piso    <= '1';
                            sel_Rfe_rd       <= '1';
                            mask_ready       <= '1';
                            Rfe_rd_from_ctrl <= '1';
                            count2_next      <= (others => '0');
                            count_next  <= count_r + 1;
                            count3_next <= count3_r + 1;
                        else
                            count_next  <= count_r + 1;
                            state_next  <= S_DECRYPT_5_WAIT;
                        end if;
                    else
                        if (count_r >= DEC_MSG_LEN_32-1) then
                            odcud_en_mreg <= '1';
                            if msg_len_r(1 downto 0) = "00" then
                                count_next <= (others => '0');
                                state_next <= S_DECRYPT_6;
                            else
                                if (count_r = DEC_MSG_LEN_32) then
                                    odcud_sel_cm <= '1';
                                    count_next <= (others => '0');
                                    state_next <= S_DECRYPT_6;
                                else
                                    count_next    <= count_r + 1;
                                end if;
                            end if;
                        else
                            odcud_en_mreg <= '1';
                            odcud_en_piso <= '1';
                            count2_next   <= count2_r + 1;
                            count_next    <= count_r + 1;
                        end if;
                    end if;
                end if;

            when S_DECRYPT_5_WAIT =>
                if (mask_valid = '1') and (R_fe_valid_to_ctrl = '1') then
                    odcud_en_mreg    <= '1';
                    odcud_en_piso    <= '1';
                    odcud_ld_piso    <= '1';
                    sel_Rfe_rd       <= '1';
                    mask_ready       <= '1';
                    Rfe_rd_from_ctrl <= '1';
                    count2_next      <= (others => '0');
                    count3_next      <= count3_r + 1;
                    state_next <= S_DECRYPT_5;
                end if;

            when S_DECRYPT_6 =>
                case msg_len_r(1 downto 0) is
                    when "00" =>
                        if (odcud_check2(2 downto 0)="111")then
                            if (p0Len_32 = 0) then
                                state_next <= S_DECRYPT_9;
                            else
                                state_next <= S_DECRYPT_7;
                            end if;
                        else
                            error      <= '1';
                            state_next <= S_RESET;
                        end if;
                    when "01" =>
                        if (odcud_check2(1 downto 0)="11")then
                            if (p0Len_32 = 0) then
                                state_next <= S_DECRYPT_9;
                            else
                                state_next <= S_DECRYPT_7;
                            end if;
                        else
                            error      <= '1';
                            state_next <= S_RESET;
                        end if;
                    when "10" =>
                        if (odcud_check2(0)='1')then
                            if (p0Len_32 = 0) then
                                state_next <= S_DECRYPT_9;
                            else
                                state_next <= S_DECRYPT_7;
                            end if;
                        else
                            error      <= '1';
                            state_next <= S_RESET;
                        end if;
                    when others =>
                        if (p0Len_32 = 0) then
                            state_next <= S_DECRYPT_9;
                        else
                            state_next <= S_DECRYPT_7;
                        end if;
                end case;

            when S_DECRYPT_7 =>
                --! CHECK_2
                if (count2_r = 2) then
                    if (mask_valid = '1') and (R_fe_valid_to_ctrl = '1')then
                        odcud_en_piso    <= '1';
                        odcud_ld_piso    <= '1';
                        sel_Rfe_rd       <= '1';
                        mask_ready       <= '1';
                        Rfe_rd_from_ctrl <= '1';
                        count2_next      <= (others => '0');
                        state_next  <= S_DECRYPT_8;
                    end if;
                else
                    odcud_en_piso <= '1';
                    count2_next   <= count2_r + 1;
                    state_next  <= S_DECRYPT_8;
                end if;

            when S_DECRYPT_8 =>
                --! CHECK_2
                if (count2_r = 2) then
                    if (mask_valid = '1') and (R_fe_valid_to_ctrl = '1')then
                        odcud_en_piso    <= '1';
                        odcud_ld_piso    <= '1';
                        sel_Rfe_rd       <= '1';
                        mask_ready       <= '1';
                        Rfe_rd_from_ctrl <= '1';
                        count2_next      <= (others => '0');
                        count_next  <= count_r + 1;
                    end if;
                    if (odcud_check2 /= "1111") then
                        error <= '1';
                        state_next  <= S_RESET;
                    else
                        if (count_r = p0Len_32-1) then
                            state_next  <= S_DECRYPT_9;
                        end if;
                    end if;
                else
                    odcud_en_piso <= '1';
                    count2_next   <= count2_r + 1;
                    count_next    <= count_r + 1;
                    if (odcud_check2 /= "1111") then
                        error <= '1';
                        state_next  <= S_RESET;
                    else
                        if (count_r = p0Len_32-1) then
                            state_next  <= S_DECRYPT_9;
                        end if;
                    end if;
                end if;

            when S_DECRYPT_9 =>
                bpgm_mgf_sel_next <= '0';
                r_F_sel_next      <= '0';
                count_next        <= (others => '0');
                state_next        <= S_DECRYPT_10;

            when S_DECRYPT_10 =>
                bpgm_mgf_start  <= '1';
                poly_mult_start <= '1';
                state_next      <= S_DECRYPT_11;

            when S_DECRYPT_11 =>
                if (chk_failed_bwc1 = '1') then
                    error      <= '1';
                    state_next <= S_RESET;
                elsif (poly_mult_done = '1') then
                    poly_mult_start <= '1';
                    state_next      <= S_DECRYPT_12;
                end if;

            when S_DECRYPT_12 =>
                --! Check 3
                sel_Rfe_rd  <= '1';
                if (R_fe_valid_to_ctrl = '1') then
                    Rfe_rd_from_ctrl <= '1';
                    if (check_3 = '1') then
                        if (count_r = NUM_704_Blk-1) then
                            count_next    <= (others => '0');
                            state_next    <= S_DECRYPT_13;
                        else
                            count_next    <= count_r + 1;
                        end if;
                    else
                        error      <= '1';
                        state_next <= S_RESET;
                    end if;
                end if;

            when S_DECRYPT_13 =>
                --! send header
                ctrl_do_valid  <= '1';
                sel_odcu       <= '1';
                sel_m_cm_addr  <= '1';
                sel_ctrl_header<= '1';
                m_cm_addr_ctrl <= count_r(5 downto 0);
                ctrl_dout <= MESSAGE & "00011" & X"00" & X"00" & msg_len_r;
                if (ctrl_do_ready = '1') then
                    count_next  <= count_r + 1;
                    state_next  <= S_DECRYPT_14;
                end if;

            when S_DECRYPT_14 =>
                ctrl_do_valid  <= '1';
                sel_odcu       <= '1';
                sel_m_cm_addr  <= '1';
                if (ctrl_do_ready = '1') then
                    m_cm_addr_ctrl <= count_r(5 downto 0);
                    m_cm_rd_ctrl   <= '1';
                    if (count_r = DEC_MSG_LEN_32_CEIL) then
                        count_next   <= (others => '0');
                        state_next   <= S_DECRYPT_15;
                    else
                        count_next    <= count_r + 1;
                    end if;
                else
                    m_cm_addr_ctrl <= count_r(5 downto 0) - 1;
                    count_next     <= count_r - 1;
                    state_next     <= S_DECRYPT_14_WAIT;
                end if;

            when S_DECRYPT_14_WAIT =>
                ctrl_do_valid  <= '1';
                sel_odcu       <= '1';
                sel_m_cm_addr  <= '1';
                if (ctrl_do_ready = '1') then
                    m_cm_addr_ctrl <= count_r(5 downto 0)+1;
                    m_cm_rd_ctrl   <= '1';
                    if (count_r = DEC_MSG_LEN_32_CEIL-1) then
                        count_next   <= (others => '0');
                        state_next   <= S_DECRYPT_15;
                    else
                        count_next   <= count_r + 2;
                        state_next   <= S_DECRYPT_14;
                    end if;
                else
                    m_cm_addr_ctrl <= count_r(5 downto 0);
                end if;

            when S_DECRYPT_15 =>
                --! send the status
                ctrl_do_valid  <= '1';
                sel_odcu       <= '1';
                sel_ctrl_header<= '1';
                ctrl_dout      <=  SUCCESS & X"0000000";
                if (ctrl_do_ready = '1') then
                    state_next  <= S_INST_SEL;
                end if;

            when others =>
                state_next <= S_RESET;



        end case;
    end process;

end behavioral;
